/*********************************************************************
*               SEGGER MICROCONTROLLER SYSTEME GmbH                  *
*       Solutions for real time microcontroller applications         *
**********************************************************************
*                                                                    *
*       (C) 2012   SEGGER Microcontroller Systeme GmbH               *
*                                                                    *
*       www.segger.com     Support: support@segger.com               *
*                                                                    *
**********************************************************************
*                                                                    *
*       J-Link * programming tool for Flash microcontrollers         *
*                                                                    *
**********************************************************************

File    : FLASH.C
Purpose : Support for programming and erasing CPUs internal Flash

--------  END-OF-HEADER  ---------------------------------------------
*/

#include "MAIN.H"
#include <string.h>

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

#define FLASH_BASEADDR  (0x08000000)

#define IWDG            ((IWDG_REGS*) (0x40003000))
#define FLASH           ((FLASH_REGS*)(0x40023C00))

#define FLASH_KEY1      (0x45670123)
#define FLASH_KEY2      (0xCDEF89AB)

#define __RAMFUNC       __attribute__((section(".data")))

/*********************************************************************
*
*       Types
*
**********************************************************************
*/

typedef struct {
  U32 acr;
} RESTORE_INFO;

typedef union {
  struct {
    void* pSrc;
    void* pDest;
    U32   NumHWords;
    U32   WritePara0;
    U32   WritePara1;
    U32   WritePara2;
  } In;
  struct {
    void* FailAddr;
  } Out;
} IO_PARA;

//
// Independent watchdog registers
//
typedef struct {
  volatile U32 KR;
  volatile U32 PR;
  volatile U32 RLR;
  volatile U32 SR;
} IWDG_REGS;

//
// Flash registers
//
typedef struct {
  volatile U32 ACR;
  volatile U32 KEYR;
  volatile U32 OPTKEYR;
  volatile U32 SR;
  volatile U32 CR;
  volatile U32 OPTCR;
} FLASH_REGS;

/*********************************************************************
*
*       Static code
*
**********************************************************************
*/
/*********************************************************************
*
*       _AddrToSectorIndex
*/
static U32 __RAMFUNC _AddrToSectorIndex(U32 Addr) {
  Addr &= 0xFFFFFF;
  if (Addr >= 0x20000) {
    return ((Addr - 0x20000) / 0x20000) + 5;
  } else if (Addr >= 0x10000) {
    return 4;
  } else {
    return (Addr / 0x4000);
  }
}

/*********************************************************************
*
*       _SectorIndexToSectorOffset
*/
static U32 __RAMFUNC _SectorIndexToSectorOffset(U32 SectorIndex) {
  if (SectorIndex >= 5) {
    return 0x20000 + ((SectorIndex - 5) * 0x20000);
  } else if (SectorIndex >= 4) {
    return 0x10000;
  } else {
    return SectorIndex * 0x4000;
  }
}

/*********************************************************************
*
*       _IsSectorBaseAddr
*
*  Function description
*    This routine checks if Addr is the base addr of a sector
*/
static int __RAMFUNC _IsSectorBaseAddr(U32 Addr) {
  U32 SectorIndex;
  U32 SectorOffset;

  SectorIndex  = _AddrToSectorIndex(Addr);
  SectorOffset = _SectorIndexToSectorOffset(SectorIndex);
  if (SectorOffset == (Addr & 0xFFFFFF)) {
    return 1;   // Is sector base addr
  }
  return 0;     // No sector base addr
}

/*********************************************************************
*
*       _FeedWatchdog
*/
static void __RAMFUNC _FeedWatchdog(void) {
  IWDG->KR = 0xAAAA;      // Feed watchdog
}

/*********************************************************************
*
*       _Prepare
*/
static void __RAMFUNC _Prepare(void* pRestoreInfo) {
  volatile RESTORE_INFO * pInfo;
  U32 v;

  pInfo = (volatile RESTORE_INFO *)pRestoreInfo;
  pInfo->acr = FLASH->ACR & 0x00001F07;
  FLASH->ACR = 7; // Disable all caches and set wait-states to 7 to be on the safe side
  do {
    v = FLASH->ACR;
    if ((v & 0x00001F07) == 7) {  // Wait until write has completely succeeded
      break;
    }
  } while (1);
  FLASH->ACR = (1 << 11) | (1 << 12) | 7;  // Reset data and instruction cache
  FLASH->ACR = 7;                          // Invalidate reset of data and instruction cache
}

/*********************************************************************
*
*       _Restore
*/
static void __RAMFUNC _Restore(void* pRestoreInfo) {
  volatile RESTORE_INFO * pInfo;
  U32 v;
  //
  // Reset data and instruction cache
  //
  FLASH->ACR = (1 << 11) | (1 << 12) | 7;  // Reset data and instruction cache
  FLASH->ACR = 7;  // Invalidate reset of data and instruction cache
  //
  // Restore cache settings
  //
  pInfo = (volatile RESTORE_INFO *)pRestoreInfo;
  FLASH->ACR = pInfo->acr; // Restore flash wait states and cache settings
  do {
    v = FLASH->ACR;
    if ((v & 0x00001F07) == pInfo->acr) {  // Wait until write has completely succeeded
      break;
    }
  } while (1);
}

/*********************************************************************
*
*       _GetSectorSize
*
*  Function description
*    Get size of current sector in bytes
*/
static U32 __RAMFUNC _GetSectorSize(U32 SectorAddr) {

  if (SectorAddr >= 0x08020000) {
    return 0x20000;
  } else if (SectorAddr >= 0x08010000) {
    return 0x10000;
  } else {
    return 0x4000;
  }
}

/*********************************************************************
*
*       _CheckIfBlank
*
*  Function description
*   Check if a given memory area is blank/erased.
*
*  Return value
*    0  Sector is not blank
*    1  Sector is blank
*/
static int __RAMFUNC _CheckIfBlank(U32 Addr, U32 NumBytes) {
  U32 * pData;
  int WDTCnt;

  WDTCnt = 0;
  pData = (U32*)Addr;
  _FeedWatchdog();
  do {
    if (*pData++ != 0xFFFFFFFF) {     // Erased state of flash is 0xFFFFFFFF
      return 0;
    }
    if (++WDTCnt == 256) {
      _FeedWatchdog();
      WDTCnt = 0;
    }
    NumBytes -= 4;
  } while(NumBytes);
  return 1;
}

/*********************************************************************
*
*       _Program
*
* Purpose:
*   This function programs the flash memory in units of 16-bits.
*
* Parameters:
*   p->In.pSrc              Source address (data to be programmed)
*   p->In.pDest             Destination address
*   p->In.NumItems          Specifies the number of 16-bit items. Value is always a multiple of 2.
*   p->In.WritePara0        (optional, not used by all flashes)
*   p->In.WritePara1        (optional, not used by all flashes)
*   p->In.WritePara2        (optional, not used by all flashes)
*
*   p->Out.FailAddr         Contains address of item that caused an error (only used in case of error)
*
* Return values:
*    0 = Operation completed successfully
*    1 = Block verification error (Detected by generic RAMCode, do not use!)
*    2 = Item verification error
*    3 = Timeout occurred
*    4 = Program error
*    5 = Program 1 over 0
*    6 = Sector is locked
*    7 = Erase error
*    8 = No flash memory
*   >8 = Unspecified error (reserved for further error codes)
*   <0 = Algo depending error code (usually a small integer such as -1, -2, -3)
*/
static int __RAMFUNC _Program(IO_PARA* p) {
  volatile U32* pSrc;
  volatile U32* pDest;
  U32 Status;
  U32 NumWords;   // Will always be a multiple of 4 since a item which shall be programmed must not cross a 128-bit region
  int r;

  r        = 0;
  NumWords = p->In.NumHWords >> 1;
  pSrc     = (volatile U32*)p->In.pSrc;
  pDest    = (volatile U32*)p->In.pDest;
  //
  // Clear all error bits
  //
  FLASH->SR = 0
            | (1 << 1)   // Clear OPERR  bit by writing 1 to it
            | (1 << 4)   // Clear WRPERR bit by writing 1 to it
            | (1 << 5)   // Clear PGAERR bit by writing 1 to it
            | (1 << 6)   // Clear PGPERR bit by writing 1 to it
            | (1 << 7)   // Clear PGSERR bit by writing 1 to it
            ;
  //
  // Unlock the flash control register
  //
  FLASH->KEYR = FLASH_KEY1;
  FLASH->KEYR = FLASH_KEY2;
  //
  // Setup flash controller
  //
  FLASH->CR  = (2 << 8);
  FLASH->CR |= (1 << 0);
  //
  // Program items
  //
  do {
    if (*pSrc != 0xFFFFFFFF) {    // Skip blank words when programming
      *pDest++ = *pSrc++;
      //
      // Wait for operation to complete
      //
      do {
       _FeedWatchdog();
       Status = FLASH->SR;
      } while (Status & (1uL << 16));
      if (Status & 0xFE) {
        r = 4;
        goto Done;
      }
    } else {
      pSrc++;
      pDest++;
    }
  } while (--NumWords);
Done:
  FLASH->CR = (1uL << 31); // Lock access to flash control register. Not entirely sure why this is needed, but it seems if we do not lock the flash control register again, flash programming sometimes fails..
  //
  // Clear all error bits
  //
  FLASH->SR = 0
            | (1 << 1)   // Clear OPERR  bit by writing 1 to it
            | (1 << 4)   // Clear WRPERR bit by writing 1 to it
            | (1 << 5)   // Clear PGAERR bit by writing 1 to it
            | (1 << 6)   // Clear PGPERR bit by writing 1 to it
            | (1 << 7)   // Clear PGSERR bit by writing 1 to it
            ;
  return r;
}

/*********************************************************************
*
*       _Erase
*
* Purpose:
*   This function erases one or more flash sectors
*   or performs a chip erase.
*
* Parameters:
*   NumSectors              NumSectors (0 means chip erase)
*   ErasePara0              (optional, not used by all flashes)
*   ErasePara1              (optional, not used by all flashes)
*   ErasePara2              (optional, not used by all flashes)
*
* Return values:
*    0 = Operation completed successfully
*    1 = Block verification error (Detected by generic RAMCode, do not use!)
*    2 = Item verification error
*    3 = Timeout occurred
*    4 = Program error
*    5 = Program 1 over 0
*    6 = Sector is locked
*    7 = Erase error
*    8 = No flash memory
*   >8 = Unspecified error (reserved for further error codes)
*   <0 = Algo depending error code (usually a small integer such as -1, -2, -3)
*/
static int __RAMFUNC _Erase(U32 NumSectors, U32 BaseAddr, U32 SectorIndex, U32 ErasePara2) {
  U32 Status;
  U32 NumSectorsRem;
  U32 SectorAddr;
  U32 SectorSize;
  int r;

  (void)BaseAddr;
  (void)ErasePara2;

  if (NumSectors == 0) {
    NumSectorsRem = 1;
  } else {
    NumSectorsRem = NumSectors;
  }
  r = 0;
  do {
    //
    // Clear all error bits
    //
    FLASH->SR = 0
              | (1 << 1)   // Clear OPERR  bit by writing 1 to it
              | (1 << 4)   // Clear WRPERR bit by writing 1 to it
              | (1 << 5)   // Clear PGAERR bit by writing 1 to it
              | (1 << 6)   // Clear PGPERR bit by writing 1 to it
              | (1 << 7)   // Clear PGSERR bit by writing 1 to it
              ;
    if (NumSectors) {
      SectorAddr = _SectorIndexToSectorOffset(SectorIndex) + 0x08000000;
      SectorSize = _GetSectorSize(SectorAddr);
      if (_CheckIfBlank(SectorAddr, SectorSize)) {
        goto SkipErase;
      }
    }
    //
    // Unlock the flash control register
    //
    FLASH->KEYR = FLASH_KEY1;
    FLASH->KEYR = FLASH_KEY2;
    //
    // Setup flash controller
    //
    if (NumSectors == 0) {         // 0 means mass erase
      FLASH->CR   = (1uL << 2);
    } else {
      FLASH->CR   = (1uL << 1) | ((SectorIndex & 0xF) << 3);    // Set flash controller to sector erase mode and select sector to be erased
    }
    FLASH->CR  |= (1uL << 16);   // Start erase operation
    //
    // Wait for operation to complete
    //
    //  Status = *((volatile U32 *)(0x20000000));
    //  Status = *((volatile U32 *)(0x20000000));
    do {
      _FeedWatchdog();
      Status = FLASH->SR;
    } while (Status & (1uL << 16));
    FLASH->CR = (1uL << 31);        // Lock access to flash control register. Not entirely sure why this is needed, but it seems if we do not lock the flash control register again, flash programming sometimes fails..
SkipErase:
    SectorIndex++;
  } while(--NumSectorsRem);
  if (Status & 0xFE) {
    //
    // Clear all error bits
    //
    FLASH->SR = 0
              | (1 << 1)   // Clear OPERR  bit by writing 1 to it
              | (1 << 4)   // Clear WRPERR bit by writing 1 to it
              | (1 << 5)   // Clear PGAERR bit by writing 1 to it
              | (1 << 6)   // Clear PGPERR bit by writing 1 to it
              | (1 << 7)   // Clear PGSERR bit by writing 1 to it
              ;
    r = 7;
  }
  return r;
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/
/*********************************************************************
*
*       FLASH_WriteEx
*
* Function description
*   This routine writes data into any section of the flash.
*   This routine will also erase the sector if the start addr. is the
*   start addr. of a sector.
*
*  Notes
*    (1)  Source data must be 4 byte aligned
*    (2)  Dest ptr must be page aligned
*    (3)  NumBytes must be a multiple of the page size
*/
void __RAMFUNC FLASH_WriteEx(U32 * pDest, const U32 * pSrc, int NumBytes, int AutoErase) {
  RESTORE_INFO Info;
  IO_PARA Paras;
  U32 StartSectorNumber;
  U32 EndSectorNumber;
  U32 SectorNumber;

  (void)pDest;
  (void)pSrc;
  (void)NumBytes;
  (void)AutoErase;

  _Prepare(&Info);
  if (AutoErase) {
    if (_IsSectorBaseAddr((U32)pDest)) {
      StartSectorNumber = _AddrToSectorIndex((U32)pDest);
      EndSectorNumber   = _AddrToSectorIndex((U32)((U8*)pDest + NumBytes - 1));
      for (SectorNumber = StartSectorNumber; SectorNumber <= EndSectorNumber; SectorNumber++) {
        _Erase(1, FLASH_BASEADDR, SectorNumber, 0);
      }
    }
  }
  Paras.In.pSrc       = (void*)pSrc;
  Paras.In.pDest      = (void*)pDest;
  Paras.In.NumHWords  = NumBytes >> 1;
  Paras.In.WritePara0 = FLASH_BASEADDR;
  _Program(&Paras);
  _Restore(&Info);
  do {
    if (*pSrc++ != *pDest++) {
      while(1);
    }
    NumBytes -= 4;
  } while (NumBytes);
}

/*********************************************************************
*
*       FLASH_Write
*/
void __RAMFUNC FLASH_Write(U32 * pDest, const U32 * pSrc, int NumBytes) {
  FLASH_WriteEx(pDest, pSrc, NumBytes, 1);
}

/*********************************************************************
*
*       FLASH_EraseSector
*
* Function description
*   Erases a complete sector
*
*  Parameters
*    p    Pointer to any address in the sector
*/
void __RAMFUNC FLASH_EraseSector(U32* p) {
  U32 SectorIndex;
  RESTORE_INFO Info;

  _Prepare(&Info);
  SectorIndex = _AddrToSectorIndex((U32)p);
  _Erase(1, FLASH_BASEADDR, SectorIndex, 0);
  _Restore(&Info);
}

/*********************************************************************
*
*       FLASH_ClrSecurityBit
*/
void __RAMFUNC FLASH_ClrSecurityBit(void) {
}

/*********************************************************************
*
*       FLASH_SetSecurityBit
*
*  Function description
*    Activates read protection of flash memory to secure J-Link against read-out.
*/
void __RAMFUNC FLASH_SetSecurityBit(void) {
  U32 v;

  v = FLASH->OPTCR;
  if (((v >> 8) & 0xFF) == 0xAA) {    // Flash unprotected?
    if (v & 1) {
      FLASH->OPTKEYR = 0x08192A3B;
      FLASH->OPTKEYR = 0x4C5D6E7F;
      v &= ~(1uL);
    }
    v |= (0xFF << 8);   // Read protection level 1 since this level is reversible so we can reprogram J-Links if necessary
    FLASH->OPTCR = v;
    //
    // Start option byte programming
    //
    FLASH->OPTCR = v | (1 << 1);
    //
    // Wait for operation to complete
    //
    do {
      v = FLASH->SR;
    } while ((v >> 16) & 1);           // Wait for BSY bit to be cleared
  }
}

/*********************************************************************
*
*       FLASH_GetSecurityBit
*/
U32 __RAMFUNC FLASH_GetSecurityBit(void) {
  return 0;
}

/*************************** End of file ****************************/
